/*
  Utility class for handling the layout of 2D elements contained in a 2D canvas.

  Very useful to specify the design of 2D widgets within a screen whose resolution can vary.
  Supports anchoring to the sides, giving coordinates as a factor or percentage of the canvas,
  centering the resulting 2D element, and even specifying one dimension's size depending on the
  other one, which is good to maintain aspect ratios.

  Whenever the dimensions of the canvas change, call CRectLayout::Calc() on your Layout instances
  before rendering, and they will realign themselves based on your static specification.

  Special thanks to Cesar Monterde for his input and previous implementation.

  Copyright (C) 2006 by Javier Arvalo (http://www.iguanademos.com/Jare)

This software is provided 'as-is', without any express or implied warranty.
In no event will the authors be held liable for any damages arising from the use of this software.

Permission is granted to anyone to use this software for any purpose, including commercial applications,
and to alter it and redistribute it freely, subject to the following restrictions:

 1 - The origin of this software must not be misrepresented. You must not claim that you wrote
     the original software. If you use this software in a product, an acknowledgment in the product
     documentation would be appreciated but is not required.
 2 - Altered source versions must be plainly marked as such, and must not be misrepresented as being
     the original software.
 3 - This notice may not be removed or altered from any source distribution.
*/

#ifndef LAYOUT_H_
#define LAYOUT_H_

#pragma once

// ------------------------------------------------------------------------
// Forward declaration
// ------------------------------------------------------------------------
template<class T, int FACTOR_RANGE = 1>
class CRectLayoutDesc_t;

// ------------------------------------------------------------------------
// Manage the layout of a segment (offset, length) on a single dimension
// ------------------------------------------------------------------------
template<class T, int FACTOR_RANGE = 1>
class CSegLayoutDesc_t
{
public:
  enum
  {
    F_SET    = 0x00,      // Absolute offset
    F_FACTOR = 0x01,      // Factor of the parent's length

    F_END    = 0x10,      // Offset between endpoints rather than between origins
    F_CENTER = 0x20,      // Only for pos - center the resulting segment around the position calculated
  };

  CSegLayoutDesc_t(): o(0), l(0), to(F_SET), tl(F_SET) { }
  CSegLayoutDesc_t(T _o, T _l,
                   unsigned _to = F_SET, unsigned _tl = F_SET)
    : o(_o), to(_to), l(_l), tl(_tl) { }

  void SetOffset(T _o, unsigned _to = F_SET)                          { o = _o; to = _to; }
  void SetLength(T _l, unsigned _tl = F_SET)                          { l = _l; tl = _tl; }
  void SetSeg(T _o, T _l, unsigned _to = F_SET, unsigned _tl = F_SET) { o = _o; to = _to; l = _l; tl = _tl; }

  // ------------------------------------
  // Perform calculations

  // Calculate offset and length
  void Calc(T po, T pl, T *pro, T *prl) const
  {
    T rl = CalcLength(po, pl);
    if (pro) *pro = CalcOffset(po, pl, rl);
    if (prl) *prl = rl;
  }

  // Calc offset with already-known length
  T CalcOffset(T po, T pl, T rl) const
  {
    T ro;
    if (to & F_FACTOR)
      ro = ((to & F_END)? pl*(T(FACTOR_RANGE)-o) - rl : pl*o)/T(FACTOR_RANGE);
    else
      ro = (to & F_END)? pl - rl - o : o;
    if (to & F_CENTER)
      ro -= (to & F_END)? -rl/2 : rl/2;

    return po + ro;
  }

  // Calc length
  T CalcLength(T po, T pl) const
  {
    T rl;
    if (tl & F_FACTOR)
      rl = pl*((tl & F_END)? T(FACTOR_RANGE)-l : l)/T(FACTOR_RANGE);
    else
      rl = (tl & F_END)? pl - l : l;
    return rl;
  }

protected:

  T o, l;
  unsigned to, tl;

  friend class CRectLayoutDesc_t<T, FACTOR_RANGE>;
};

// -----------------------------------------------------------------
// -----------------------------------------------------------------

// ------------------------------------------------------------------------
// Manage the layout of a box in two dimensions
// ------------------------------------------------------------------------
template<class T, int FACTOR_RANGE>
class CRectLayoutDesc_t
{
public:
  enum
  {
    // CSegLayoutDesc_t:: constants repeated for convenience
    F_SET    = 0x00,      // Absolute offset
    F_FACTOR = 0x01,      // Factor of the parent's range

    F_END    = 0x10,      // Offset between endpoints rather than between origins
    F_CENTER = 0x20,      // Only for pos - center the resulting segment around the position calculated
    F_OTHER  = 0x40,      // Only for size - value is relative to the OTHER dimension's size
  };

  CRectLayoutDesc_t (): x(), y() { }
  CRectLayoutDesc_t (T _x, T _y, T _w, T _h,
                     unsigned _tx = F_SET, unsigned _ty = F_SET, unsigned _tw = F_SET, unsigned _th = F_SET)
    : x(_x, _w, _tx, _tw), y(_y, _h, _ty, _th) { }

  void SetX(T _x, unsigned _tx = F_SET) { x.SetOffset(_x, _tx); }
  void SetY(T _y, unsigned _ty = F_SET) { y.SetOffset(_y, _ty); }
  void SetW(T _w, unsigned _tw = F_SET) { x.SetLength(_w, _tw); }
  void SetH(T _h, unsigned _th = F_SET) { y.SetLength(_h, _th); }
  void Set(T _x, T _y, T _w, T _h,
           unsigned _tx = F_SET, unsigned _ty = F_SET, unsigned _tw = F_SET, unsigned _th = F_SET)
  {
    x.SetSeg(_x, _w, _tx, _tw);
    y.SetSeg(_y, _h, _ty, _th);
  }

  // ------------------------------------
  // Perform calculations

  // Calculate offsets and lengths
  void Calc(T px, T py, T pw, T ph, T *prx, T *pry, T *prw, T *prh) const
  {
    T rx;
    T ry;
    T rw;
    T rh;

    // Calc the dimension that is not dependent on the other
    if (x.tl & F_OTHER)
      y.Calc(py, ph, &ry, &rh);
    else
      x.Calc(px, pw, &rx, &rw);

    // Calculate the other dimension, using the previous result if appropriate
    if (x.tl & F_OTHER)
    {
      rw = x.CalcLength(px, rh);       // Width is relative to the height
      rx = x.CalcOffset(px, pw, rw);
    }
    else if (y.tl & F_OTHER)
    {
      rh = y.CalcLength(py, rw);       // Height is relative to the width
      ry = y.CalcOffset(py, ph, rh);
    }
    else
      y.Calc(py, ph, &ry, &rh);   // Normal calculation

    if (prx) *prx = rx;
    if (pry) *pry = ry;
    if (prw) *prw = rw;
    if (prw) *prh = rh;
  }
protected:

  CSegLayoutDesc_t<T, FACTOR_RANGE> x, y;
};

// ------------------------------------------------------------------------
// Store a layout and the calculated rectangle
// ------------------------------------------------------------------------
template<class T, int FACTOR_RANGE = 1>
class CRectLayout_t: public CRectLayoutDesc_t<T, FACTOR_RANGE>
{
public:
  T x, y, w, h;

  CRectLayout_t (): x(), y(), w(), h() { }
  CRectLayout_t (T _x, T _y, T _w, T _h,
                 unsigned _tx = F_SET, unsigned _ty = F_SET, unsigned _tw = F_SET, unsigned _th = F_SET)
    : CRectLayoutDesc_t(_x, _y, _w, _h, _tx, _ty, _tw, _th), x(), y(), w(), h() { }

  // ------------------------------------
  // Perform calculations

  // Calculate the already set values for a given parent frame
  void Calc(T px, T py, T pw, T ph)
  {
    CRectLayoutDesc_t::Calc(px, py, pw, ph, &x, &y, &w, &h);
  }

  // Initialize and calculate the values for a given parent frame
  void Setup(T px, T py, T pw, T ph,
             T _x, T _y, T _w, T _h,
             unsigned _tx = F_SET, unsigned _ty = F_SET, unsigned _tw = F_SET, unsigned _th = F_SET)
  {
    Set(_x, _y, _w, _h, _tx, _ty, _tw, _th);
    Calc(px, py, pw, ph);
  }

  // ------------------------------------
  // Heh why not, utility function to hit-test the calculated rectangle
  bool inside(T px, T py) const { return px >= x && px < x+w && py >= y && py < y+h; }
};

// ------------------------------------------------------------------------
// Convenient typedefs
// ------------------------------------------------------------------------
typedef CRectLayout_t<float, 1> CRectLayout;    // Factor for F_FACTOR values for floats is 1.f
typedef CRectLayout_t<int, 100> CRectLayoutI;   // Factor for F_FACTOR values for ints is 100

/*
  Example usages:

  CRectLayout l1, l2, l3;

  // One-time initialization
  l1.Set(0, 0, 500.f, .25f,
        CRectLayout::F_SET,                             // Anchored to the left side
        CRectLayout::F_END,                             // Anchored to the bottom
        CRectLayout::F_SET,                             // Width is absolute 500
        CRectLayout::F_FACTOR | CRectLayout::F_OTHER ); // Height is 25% of the width (i.e. 125)

  l2.Set(.5f, 50.f, .4f, .25f,
        CRectLayout::F_FACTOR | CRectLayout::F_CENTER,  // Centered horizontally around the center of the canvas
        CRectLayout::F_SET,                             // Anchored to the 50 below the top
        CRectLayout::F_FACTOR,                          // Width is 40% of the canvas' width
        CRectLayout::F_FACTOR | CRectLayout::F_OTHER ); // Height is 25% of the calculated width

  // Calculate the actual rectangles
  l1.Calc(0, 0, 640, 480);                              // Compute the actual rect for a canvas of 640x480
                                                        // Rect is contained in the x,y,w,h members
  l2.Calc(l1.x, l1.y, l1.w, l1.h);                      // Compute the actual rect relative to l1's Rect

  l3.Setup(0, 0, 1024, 768,
           .5f, .25f, 4.f, .1f,                         // Initialize and calculate in one go from 1024x768 canvas
        CRectLayout::F_FACTOR | CRectLayout::F_CENTER,  // Centered horizontally around the center of the canvas
        CRectLayout::F_FACTOR | CRectLayout::F_CENTER,  // Centered vertically around the 1/4 of the canvas
        CRectLayout::F_FACTOR | CRectLayout::F_OTHER ,  // Width is 400% of the calculated height
        CRectLayout::F_FACTOR);                         // Height is 10% of the canvas' height
*/

#endif // LAYOUT_H_
